package at.ac.tuwien.dbai.ges.instances;

import at.ac.tuwien.dbai.ges.instances.breaks.BreakGenerator;
import at.ac.tuwien.dbai.ges.instances.employee.*;
import at.ac.tuwien.dbai.ges.instances.restriction.*;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftDetails;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftGenerator;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftRow;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftRowGenerator;
import at.ac.tuwien.dbai.ges.instances.tasks.TaskGenerator;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.IntStream;

/**
 * Generates the schedule.
 *
 * @author Lucas Kletzander
 */
public class ScheduleGenerator {

    private final ScheduleConfig config;
    private ShiftGenerator shiftGenerator;
    private List<ShiftRow> selection;
    private RestrictionMonitor restrictionMonitor;
    private EmployeeMonitor employeeMonitor;
    private ShiftRowGenerator shiftRowGenerator;
    private BreakGenerator breakGenerator;
    private List<EmployeeSchedule> data;
    private SkillGenerator skillGenerator;
    private TaskGenerator taskGenerator;

    public ScheduleGenerator(ScheduleConfig config) {
        this.config = config;
    }

    public ScheduleConfig getConfig() {
        return config;
    }

    public ShiftGenerator getShiftGenerator() {
        return shiftGenerator;
    }

    public RestrictionMonitor getRestrictionMonitor() {
        return restrictionMonitor;
    }

    public EmployeeMonitor getEmployeeMonitor() {
        return employeeMonitor;
    }

    public ShiftRowGenerator getShiftRowGenerator() {
        return shiftRowGenerator;
    }

    public BreakGenerator getBreakGenerator() {
        return breakGenerator;
    }

    public List<EmployeeSchedule> getData() {
        return data;
    }

    public SkillGenerator getSkillGenerator() {
        return skillGenerator;
    }

    public TaskGenerator getTaskGenerator() {
        return taskGenerator;
    }

    public void createShifts() {

        System.out.println("Generating shifts...");
        shiftGenerator = new ShiftGenerator(config);
        shiftGenerator.generateShifts();

        int shiftTarget = (config.getDays() + config.getHistory()) * config.getShiftTypes() * config.getShiftAvg();
        restrictionMonitor = new RestrictionMonitor(config.getContracts());

        if (config.getWeekTimeRestriction() != RestrictionType.NONE) {
            restrictionMonitor.addRestriction(new WeekTimeRestriction(config, shiftGenerator));
        }
        if (config.getTimeRestriction() != RestrictionType.NONE) {
            restrictionMonitor.addRestriction(new TimeRestriction(config, shiftGenerator));
        }
        if (config.getPatternRestriction() != RestrictionType.NONE) {
            restrictionMonitor.addRestriction(new PatternRestriction(config.getPatternRestriction() == RestrictionType.GLOBAL, shiftGenerator, config.getPatternRestrictionCount()));
        }

        long factor = restrictionMonitor.getFactor();
        long amount = factor * shiftTarget * (factor * shiftTarget < 10000 ? 10 : 1);

        System.out.println("Generating " + amount + " shift instance candidates...");

        shiftRowGenerator = new ShiftRowGenerator(config, shiftGenerator, restrictionMonitor);
        IntStream.range(0, config.getContracts()).forEach(i -> shiftRowGenerator.generateRows(i, config.getMinOn()[i], config.getMaxOn()[i], config.getMinOff()[i], config.getMaxOff()[i], amount));

        System.out.println("Starting to choose shifts...");
        restrictionMonitor.applyRestrictions();
        selection = shiftRowGenerator.chooseShiftRows(shiftTarget);

        employeeMonitor = new EmployeeMonitor();
        if (config.getMaxShiftPreferences() > 0) {
            employeeMonitor.addConfiguration(new ShiftPreference(config, shiftGenerator));
        }
        employeeMonitor.addConfiguration(new HistoryConfiguration(config, shiftGenerator));

    }

    public void createBreaks() {

        if (config.getBreakTypes() > 0) {

            System.out.println("Generating breaks...");
            breakGenerator = new BreakGenerator(config);
            breakGenerator.generateBreakTypes();
            breakGenerator.generateBreakConfigurations(shiftGenerator);
            breakGenerator.generateShiftDetails(shiftGenerator);

            data = new ArrayList<>();
            for (int i = 0; i < selection.size(); i++) {
                EmployeeSchedule schedule = new EmployeeSchedule(i, selection.get(i).getContract());
                schedule.setDetails(breakGenerator.scheduleBreaks(selection.get(i)));
                data.add(schedule);
            }

        } else {

            data = new ArrayList<>();

            for (int i = 0; i < selection.size(); i++) {

                EmployeeSchedule schedule = new EmployeeSchedule(i, selection.get(i).getContract());
                List<ShiftDetails> details = new ArrayList<>();

                for (int shift : selection.get(i).getShifts()) {

                    ShiftDetails detail = null;
                    if (shift != -1) {
                        detail = new ShiftDetails(shiftGenerator.getShiftInstances().get(shift), shift);
                    }
                    details.add(detail);

                }

                schedule.setDetails(details);
                data.add(schedule);
            }

        }

    }

    public void createTasks() {

        System.out.println("Generating skills and tasks...");
        skillGenerator = new SkillGenerator(config);
        for (EmployeeSchedule schedule : data) {
            schedule.setSkills(skillGenerator.generateEmployeeSkills());
        }

        if (config.getDemandType() == DemandType.TASKS) {
            taskGenerator = new TaskGenerator(config);
            taskGenerator.generateTasks(data);
        }

    }

}
