package at.ac.tuwien.dbai.ges.instances.breaks;

import at.ac.tuwien.dbai.ges.instances.RandomProvider;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.IntStream;

/**
 * Defines a break configuration.
 *
 * @author Lucas Kletzander
 */
public class BreakConfiguration {

    private int id;
    private int minLength;
    private int maxLength;
    private int maxWorkAfterLast = 0;
    private double minBreakTimeFraction = 1;
    private double maxBreakTimeFraction = 0;
    private List<Integer> availableTypes = new ArrayList<>();
    private boolean[] typeUnused;

    public BreakConfiguration(int types) {

        IntStream.range(0, types).forEach(i -> availableTypes.add(i));
        typeUnused = new boolean[types];

    }

    public BreakConfiguration(BreakType[] types, int id) {

        this.id = id;
        chooseTypes(types);

    }

    public BreakConfiguration(BreakType[] types, int minLength, int maxLength) {

        this.minLength = minLength;
        this.maxLength = maxLength;
        chooseTypes(types);

    }

    private void chooseTypes(BreakType[] types) {

        int selection = RandomProvider.get().nextInt(types.length) + 1;
        boolean hasNormalBreak = false;

        while (availableTypes.size() < selection || !hasNormalBreak) {

            int cur = RandomProvider.get().nextInt(types.length);
            if (!availableTypes.contains(cur)) {
                hasNormalBreak |= types[cur].getStartShiftWeight() == 0;
                availableTypes.add(cur);
            }

        }

        typeUnused = new boolean[types.length];

    }

    public int getId() {
        return id;
    }

    public int getMinLength() {
        return minLength;
    }

    public int getMaxLength() {
        return maxLength;
    }

    public void addWorkAfterLast(int work) {
        maxWorkAfterLast = Math.max(maxWorkAfterLast, work);
    }

    public int getMaxWorkAfterLast() {
        return maxWorkAfterLast;
    }

    public void addBreakTimeFraction(double fraction) {
        minBreakTimeFraction = Math.min(minBreakTimeFraction, fraction);
        maxBreakTimeFraction = Math.max(maxBreakTimeFraction, fraction);
    }

    public double getMinBreakTimeFraction() {
        return minBreakTimeFraction;
    }

    public double getMaxBreakTimeFraction() {
        return maxBreakTimeFraction;
    }

    public void setTypeUnused(int type) {
        typeUnused[type] = true;
    }

    public boolean[] getTypeUnused() {
        return typeUnused;
    }

    public List<Integer> getAvailableTypes() {
        return availableTypes;
    }

    @Override
    public String toString() {
        return "BreakConfiguration{" +
                "id=" + id +
                ", minLength=" + minLength +
                ", maxLength=" + maxLength +
                ", maxWorkAfterLast=" + maxWorkAfterLast +
                ", minBreakTimeFraction=" + minBreakTimeFraction +
                ", maxBreakTimeFraction=" + maxBreakTimeFraction +
                ", availableTypes=" + availableTypes +
                '}';
    }

}
