package at.ac.tuwien.dbai.ges.instances.breaks;

import at.ac.tuwien.dbai.ges.instances.RandomProvider;
import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;
import at.ac.tuwien.dbai.ges.instances.shift.*;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 * Generates break types, break configurations and break schedules for shifts.
 *
 * @author Lucas Kletzander
 */
public class BreakGenerator {

    private final ScheduleConfig config;
    private BreakType[] breakTypes;
    private List<BreakConfiguration> breakConfigurations;
    private List<List<ShiftDetails>> detailSelection;

    public BreakGenerator(ScheduleConfig config) {
        this.config = config;
    }

    public void generateBreakTypes() {

        breakTypes = new BreakType[config.getBreakTypes()];

        for (int i = 0; i < config.getBreakTypes(); i++) {

            BreakType breakType = new BreakType();
            generateBreakLength(breakType);
            generateShiftTimeConstraint(breakType);

            if (i > 0 && RandomProvider.get().nextDouble() < config.getLunchBreakFactor()) {
                generateStartShiftConstraint(breakType);
            }

            generateWorkTimeConstraint(breakType);

            breakTypes[i] = breakType;

        }

    }

    private void generateBreakLength(BreakType breakType) {

        breakType.setMaxLength((RandomProvider.get().nextInt(config.getMaxBreakLength()) + 1) * config.getTimeSlotLength());
        breakType.setMinLength(RandomProvider.getInBounds(config.getTimeSlotLength(), breakType.getMaxLength(), config.getTimeSlotLength()));

    }

    private void generateShiftTimeConstraint(BreakType breakType) {

        breakType.setMaxEndShift(RandomProvider.getInBounds(config.getTimeSlotLength(), config.getMaxShiftLength() * config.getShiftSlotLength() / 5, config.getTimeSlotLength()));

    }

    private void generateStartShiftConstraint(BreakType breakType) {

        breakType.setMinStartShift(RandomProvider.getInBounds(config.getMaxShiftLength() * config.getShiftSlotLength() / 10, config.getMaxShiftLength() * config.getShiftSlotLength() / 3, config.getTimeSlotLength()));
        breakType.setMaxStartShift(RandomProvider.getInBounds(breakType.getMinStartShift() + breakType.getMinLength(), config.getMaxShiftLength() * config.getShiftSlotLength() * 4 / 5, config.getTimeSlotLength()));
        breakType.setStartShiftWeight(RandomProvider.get().nextInt(10) + 1);

    }

    private void generateWorkTimeConstraint(BreakType breakType) {

        breakType.setMinWorkBefore(RandomProvider.getInBounds(config.getMinBreakDistance() * config.getTimeSlotLength(), Math.max(config.getMaxShiftLength() * config.getShiftSlotLength() / 4, config.getMinBreakDistance() * config.getTimeSlotLength()), config.getTimeSlotLength()));
        breakType.setMaxWorkBefore(RandomProvider.getInBounds(breakType.getMinWorkBefore(), Math.max(breakType.getMinWorkBefore(), config.getMaxBreakDistance() * config.getTimeSlotLength()), config.getTimeSlotLength()));

    }

    public BreakType[] getBreakTypes() {
        return breakTypes;
    }

    public void generateBreakConfigurations(ShiftGenerator shiftGenerator) {

        breakConfigurations = new ArrayList<>();

        switch (config.getBreakConfigurationType()) {

            case COMMON:
                breakConfigurations.add(new BreakConfiguration(breakTypes.length));
                break;

            case PER_SHIFT_TYPE:
                IntStream.range(0, config.getShiftTypes()).forEach(i -> breakConfigurations.add(new BreakConfiguration(breakTypes, i)));
                break;

            case PER_CONTRACT:
                IntStream.range(0, config.getContracts()).forEach(i -> breakConfigurations.add(new BreakConfiguration(breakTypes, i)));
                break;

            case PER_SHIFT_LENGTH:
                int min = shiftGenerator.getShiftInstances().stream().mapToInt(ShiftInstance::getDuration).min().orElse(0);
                int max = shiftGenerator.getShiftInstances().stream().mapToInt(ShiftInstance::getDuration).max().orElse(0);
                int diff = (max - min) / 3;
                IntStream.range(0, 3).forEach(i -> breakConfigurations.add(new BreakConfiguration(breakTypes, i == 0 ? 0 : min + i * diff, i == 2 ? Integer.MAX_VALUE : min + (i + 1) * diff - 1)));
                break;

        }

    }

    public List<BreakConfiguration> getBreakConfigurations() {
        return breakConfigurations;
    }

    public void generateShiftDetails(ShiftGenerator shiftGenerator) {

        detailSelection = new ArrayList<>();

        for (int i = 0; i < shiftGenerator.getShiftInstances().size(); i++) {

            for (int j = 0; j < (config.getBreakConfigurationType() == BreakConfigurationType.PER_CONTRACT ? config.getContracts() : 1); j++) {

                List<ShiftDetails> shiftDetails = new ArrayList<>();

                for (int k = 0; k < 100 / (1.001 - config.getBreakTimeFractionRestriction()); k++) {

                    ShiftDetails detail = new ShiftDetails(shiftGenerator.getShiftInstances().get(i), i, chooseConfig(shiftGenerator, j, i));
                    generateBreaks(detail);
                    shiftDetails.add(detail);

                }

                detailSelection.add(restrictDetails(shiftDetails));

            }

        }

    }

    private void generateBreaks(ShiftDetails details) {

        int cur = details.getStartTime();
        BreakConfiguration c = breakConfigurations.get(details.getBreakConfig());
        List<Integer> normalBreaks = c.getAvailableTypes().stream().filter(i -> breakTypes[i].getStartShiftWeight() == 0).collect(Collectors.toList());
        List<Integer> lunchBreaks = c.getAvailableTypes().stream().filter(i -> breakTypes[i].getStartShiftWeight() != 0).collect(Collectors.toList());
        int breakTime = 0;

        while (cur < details.getEndTime()) {

            int finalCur = cur;
            List<Integer> possible = lunchBreaks.stream().filter(i -> finalCur - details.getStartTime() >= breakTypes[i].getEarliestBreakBefore() && finalCur - details.getStartTime() <= breakTypes[i].getLastestBreakBefore(details.getDuration())).collect(Collectors.toList());
            int possibleLunch = possible.size();
            possible.addAll(normalBreaks.stream().filter(i -> finalCur - details.getStartTime() <= breakTypes[i].getLastestBreakBefore(details.getDuration())).collect(Collectors.toList()));

            if (possible.isEmpty()) {
                break;
            }

            int type = RandomProvider.getFromCollection(possible);
            BreakType breakType = breakTypes[type];

            int start = calculateStart(details, cur, breakType, lunchBreaks.contains(type));
            int end = calculateEnd(details, start, breakType);

            for (int i = 0; i < possibleLunch; i++) {

                if (end - details.getStartTime() > breakTypes[possible.get(i)].getLastestBreakBefore(details.getDuration())) {
                    type = possible.get(i);
                    breakType = breakTypes[type];
                    start = calculateStart(details, cur, breakType, true);
                    end = calculateEnd(details, start, breakType);
                    break;
                }

            }

            if (breakType.getStartShiftWeight() != 0) {
                lunchBreaks.remove((Integer) type);
            }

            details.getBreaks().add(new Break(type, start, end));
            breakTime += end - start;
            cur = end;

        }

        details.setBreakResults(lunchBreaks, (double) breakTime / details.getDuration());

    }

    private int calculateStart(ShiftDetails details, int cur, BreakType breakType, boolean isLunch) {

        int lower = Math.max(breakType.getMinWorkBefore(), details.getStartTime() - cur + breakType.getMinStartShift());
        int start = cur + RandomProvider.getInBounds(lower, Math.max(lower, Math.min(breakType.getMaxWorkBefore(), details.getEndTime() - cur - breakType.getMaxEndShift() - breakType.getMinLength())), config.getTimeSlotLength());
        if (isLunch && start > details.getStartTime() + breakType.getMaxStartShift()) {
            start = details.getStartTime() + breakType.getMaxStartShift();
        }

        return start;

    }

    private int calculateEnd(ShiftDetails details, int start, BreakType breakType) {

        return start + RandomProvider.getInBounds(breakType.getMinLength(), Math.max(breakType.getMinLength(), Math.min(breakType.getMaxLength(), details.getEndTime() - start - breakType.getMaxEndShift())), config.getTimeSlotLength());

    }

    private List<ShiftDetails> restrictDetails(List<ShiftDetails> details) {

        details.sort((o1, o2) -> (int) Math.signum(o1.getBreakTimeFraction() - o2.getBreakTimeFraction()));
        return details.subList((int) (details.size() * config.getBreakTimeFractionRestriction() * config.getBreakTimeFractionFocus()),
                details.size() - (int) (details.size() * config.getBreakTimeFractionRestriction() * (1 - config.getBreakTimeFractionFocus())));

    }

    public List<ShiftDetails> scheduleBreaks(ShiftRow row) {

        List<ShiftDetails> details = new ArrayList<>();

        for (int shift : row.getShifts()) {

            ShiftDetails detail = null;

            if (shift != -1) {
                detail = new ShiftDetails(RandomProvider.getFromCollection(detailSelection.get(config.getBreakConfigurationType() == BreakConfigurationType.PER_CONTRACT ? shift * config.getContracts() + row.getContract() : shift)));
                detail.setUsed(breakConfigurations.get(detail.getBreakConfig()));
            }

            details.add(detail);

        }

        return details;

    }

    private int chooseConfig(ShiftGenerator shiftGenerator, int contract, int shift) {

        switch (config.getBreakConfigurationType()) {

            case PER_SHIFT_TYPE:
                return shiftGenerator.getShiftType(shift);

            case PER_CONTRACT:
                return contract;

            case PER_SHIFT_LENGTH:
                ShiftInstance instance = shiftGenerator.getShiftInstances().get(shift);
                for (int i = 0; i < breakConfigurations.size(); i++) {
                    if (instance.getDuration() >= breakConfigurations.get(i).getMinLength() && instance.getDuration() <= breakConfigurations.get(i).getMaxLength()) {
                        return i;
                    }
                }

            default:
                return 0;

        }

    }

}
