package at.ac.tuwien.dbai.ges.instances.restriction;

import at.ac.tuwien.dbai.ges.instances.RandomProvider;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftGenerator;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftRow;
import at.ac.tuwien.dbai.ges.schema.Contracts;
import at.ac.tuwien.dbai.ges.schema.ForbiddenSequence;
import at.ac.tuwien.dbai.ges.schema.NoShift;
import at.ac.tuwien.dbai.ges.schema.ObjectFactory;

import javax.xml.bind.JAXBElement;
import java.util.*;
import java.util.stream.IntStream;

/**
 * Finds patterns of two shifts (maybe including one day off in between) to exclude.
 *
 * @author Lucas Kletzander
 */
public class PatternRestriction extends Restriction {

    private final int numPatterns;
    private final ShiftGenerator handler;
    private List<int[]> patterns;
    private Map<Long, Boolean[]> patternMatches;
    private int[][] chosenPatterns;

    public PatternRestriction(boolean global, ShiftGenerator handler, int numPatterns) {

        super(global);
        this.numPatterns = numPatterns;
        this.handler = handler;
        patterns = new ArrayList<>();

        for (int i = 0; i < handler.getShiftTypes(); i++) {

            for (int j = 0; j < 2; j++) {

                List<Integer> used = new ArrayList<>();
                if (handler.getShiftTypes() <= 3) {
                    used.add(i);
                }

                for (int k = 0; k < 10 && used.size() < handler.getShiftTypes(); k++) {

                    int[] pattern = new int[2 + j];
                    pattern[0] = i;
                    if (j == 1) {
                        pattern[1] = -1;
                    }
                    pattern[1 + j] = getNewInt(handler.getShiftTypes(), used);

                    patterns.add(pattern);

                }

            }

        }

    }

    private int getNewInt(int bound, List<Integer> used) {

        int next = RandomProvider.get().nextInt(bound);
        while (used.contains(next)) {
            next = RandomProvider.get().nextInt(bound);
        }

        used.add(next);
        return next;

    }

    @Override
    protected void initResult(int size) {
        chosenPatterns = new int[size][numPatterns];
    }

    @Override
    protected void initEvaluationData() {
        patternMatches = new HashMap<>();
    }

    @Override
    protected void evaluateRow(ShiftRow row) {

        Boolean[] matches = new Boolean[patterns.size()];
        Arrays.fill(matches, false);

        for (int i = 0; i < row.getShifts().length; i++) {

            for (int j = 0; j < patterns.size(); j++) {

                int[] pattern = patterns.get(j);
                boolean found = false;

                for (int k = 0; k < pattern.length && i + k < row.getShifts().length; k++) {

                    if (handler.getShiftType(row.getShifts()[i + k]) != pattern[k]) {
                        break;
                    }

                    if (k == pattern.length - 1) {
                        found = true;
                        break;
                    }

                }

                matches[j] |= found;

            }

        }

        patternMatches.put(row.getId(), matches);

    }

    @Override
    protected void performRestriction(int contract) {

        Integer[] count = new Integer[patterns.size()];
        Arrays.setAll(count, i -> (int) patternMatches.values().stream().filter(a -> a[i]).count());

        List<Integer> countList = new ArrayList<>(Arrays.asList(count));
        Collections.sort(countList);

        List<Integer> used = new ArrayList<>();

        int index = 0;
        for (int i = 0; i < numPatterns; i++) {

            while (countList.get(index) == 0 && index + numPatterns - i < countList.size()) {
                index++;
            }

            chosenPatterns[contract][i] = chooseIndex(count, used, countList.get(index));
            index++;

        }

    }

    private int chooseIndex(Integer[] array, List<Integer> used, int value) {

        for (int i = 0; i < array.length; i++) {
            if (array[i] == value && !used.contains(i)) {
                used.add(i);
                return i;
            }
        }

        return -1;

    }

    @Override
    protected boolean checkRemove(ShiftRow row, int contract) {

        return IntStream.range(0, numPatterns)
                .filter(i -> patternMatches.get(row.getId())[chosenPatterns[contract][i]])
                .count() > 0;

    }

    @Override
    public long getFactor() {
        return numPatterns;
    }

    @Override
    public void transform(ObjectFactory objectFactory, Contracts.Contract c, int contract) {

        Contracts.Contract.PatternConstraints p = c.getPatternConstraints();
        if (p == null) {
            p = new Contracts.Contract.PatternConstraints();
            c.setPatternConstraints(p);
        }

        List<JAXBElement<?>> list = p.getCountConstraintOrSequenceConstraintOrWeekendCount();

        for (int i = 0; i < numPatterns; i++) {

            ForbiddenSequence sequence = new ForbiddenSequence();
            List<JAXBElement<?>> pList = sequence.getNoShiftOrAnyShiftOrAnyOrNoShift();

            for (int shift : patterns.get(chosenPatterns[contract][i])) {

                if (shift == -1) {
                    pList.add(objectFactory.createPatternNoShift(new NoShift()));
                } else {
                    pList.add(objectFactory.createPatternShiftTypes("S" + shift));
                }

            }

            list.add(objectFactory.createContractsContractPatternConstraintsForbiddenSequence(sequence));

        }

    }

}
