package at.ac.tuwien.dbai.ges.instances.restriction;

import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;
import at.ac.tuwien.dbai.ges.instances.employee.WorkloadRestrictionType;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftGenerator;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftRow;
import at.ac.tuwien.dbai.ges.instances.transformer.Transformer;
import at.ac.tuwien.dbai.ges.schema.Contracts;
import at.ac.tuwien.dbai.ges.schema.ObjectFactory;
import at.ac.tuwien.dbai.ges.schema.WorkloadType;

import java.util.*;

/**
 * Restricts the total work time.
 *
 * @author Lucas Kletzander
 */
public class TimeRestriction extends Restriction {

    private final ScheduleConfig config;
    private final ShiftGenerator shiftGenerator;
    private Map<Long, Integer> timeMap;
    private int[] minTime;
    private int[] maxTime;

    public TimeRestriction(ScheduleConfig config, ShiftGenerator shiftGenerator) {

        super(false);
        this.config = config;
        this.shiftGenerator = shiftGenerator;

    }

    @Override
    protected void initResult(int size) {

        minTime = new int[size];
        maxTime = new int[size];

    }

    @Override
    protected void initEvaluationData() {
        timeMap = new HashMap<>();
    }

    @Override
    protected void evaluateRow(ShiftRow row) {

        timeMap.put(row.getId(), Arrays.stream(row.getShifts()).skip(config.getHistory()).filter(j -> j != -1)
                .map(j -> shiftGenerator.getShiftInstances().get(j).getDuration())
                .sum());

    }

    @Override
    protected void performRestriction(int contract) {

        List<Integer> timeList = new ArrayList<>(timeMap.values());
        Collections.sort(timeList);
        int position = Arrays.binarySearch(timeList.toArray(), config.getContractTimes()[contract]);
        position = position < 0 ? -position - 1 : position;
        minTime[contract] = timeList.get((int) (position * config.getTimeRestrictionFactor()));
        int maxPosition = (int) (timeList.size() - (timeList.size() - position) * config.getTimeRestrictionFactor());
        maxTime[contract] = timeList.get(maxPosition < timeList.size() ? maxPosition : timeList.size() - 1);

    }

    @Override
    protected boolean checkRemove(ShiftRow row, int contract) {

        if (timeMap.get(row.getId()) == null) {
            evaluateRow(row);
        }

        return timeMap.get(row.getId()) < minTime[contract] || timeMap.get(row.getId()) > maxTime[contract];

    }

    @Override
    public long getFactor() {
        return (long) (1 / (1.01 - config.getTimeRestrictionFactor()));
    }

    @Override
    protected void transform(ObjectFactory objectFactory, Contracts.Contract c, int contract) {

        int diff = (int) ((maxTime[contract] - minTime[contract]) * (1 - config.getWorkloadTightness()));
        if (diff == 0) {
            diff = (int) (60 * (1 - config.getWorkloadTightness()));
        }

        WorkloadType workload = new WorkloadType();

        if (config.getWorkloadRestrictionType() != WorkloadRestrictionType.MAX) {
            workload.setMin(Transformer.transformTimeSpan(Math.max(0, minTime[contract] - diff)));
        }
        if (config.getWorkloadRestrictionType() != WorkloadRestrictionType.MIN) {
            workload.setMax(Transformer.transformTimeSpan(maxTime[contract] + diff));
        }

        workload.setUnit("ShiftLength");
        c.getMinRestTimeOrMinWeekRestTimeOrWorkload().add(objectFactory.createContractsContractWorkload(workload));

    }

}
