package at.ac.tuwien.dbai.ges.instances.restriction;

import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftGenerator;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftRow;
import at.ac.tuwien.dbai.ges.schema.Contracts;
import at.ac.tuwien.dbai.ges.schema.ObjectFactory;

import java.util.*;

/**
 * Restricts the maximum weekly work time.
 *
 * @author Lucas Kletzander
 */
public class WeekTimeRestriction extends Restriction {

    private final ScheduleConfig config;
    private final ShiftGenerator shiftGenerator;
    private Map<Long, Integer> maxTimeMap;
    private int[] maxTime;

    public WeekTimeRestriction(ScheduleConfig config, ShiftGenerator shiftGenerator) {

        super(false);
        this.config = config;
        this.shiftGenerator = shiftGenerator;

    }

    @Override
    protected void initResult(int size) {
        maxTime = new int[size];
    }

    @Override
    protected void initEvaluationData() {
        maxTimeMap = new HashMap<>();
    }

    @Override
    protected void evaluateRow(ShiftRow row) {

        int min = Integer.MAX_VALUE;
        int max = 0;

        for (int i = 0; i < config.getDays() / 7; i++) {

            int cur = 0;

            for (int j = 0; j < 7 && 7 * i + j < config.getDays(); j++) {
                int shift = row.getShifts()[7 * i + j];
                cur += shift == -1 ? 0 : shiftGenerator.getShiftInstances().get(shift).getDuration();
            }

            min = Math.min(min, cur);
            max = Math.max(max, cur);

        }

        maxTimeMap.put(row.getId(), max);

    }

    @Override
    protected void performRestriction(int contract) {

        List<Integer> maxTimeList = new ArrayList<>(maxTimeMap.values());
        Collections.sort(maxTimeList);
        maxTime[contract] = maxTimeList.get((int) (maxTimeMap.size() * (1 - config.getWeekTimeRestrictionFactor())));

    }

    @Override
    protected boolean checkRemove(ShiftRow row, int contract) {

        if (maxTimeMap.get(row.getId()) == null) {
            evaluateRow(row);
        }

        return maxTimeMap.get(row.getId()) > maxTime[contract];

    }

    @Override
    public long getFactor() {
        return (long) Math.pow(1 / (1.01 - config.getWeekTimeRestrictionFactor()), 2);
    }

    @Override
    protected void transform(ObjectFactory objectFactory, Contracts.Contract c, int contract) {}

}
