package at.ac.tuwien.dbai.ges.instances.shift;

import at.ac.tuwien.dbai.ges.instances.RandomProvider;
import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.IntStream;

/**
 * Generates shift types and shift instances.
 *
 * @author Lucas Kletzander
 */
public class ShiftGenerator {

    private final ScheduleConfig config;
    private final List<ShiftInstance> shiftInstances;
    private final ShiftTypeDefinition[] definitions;
    private final List<List<Integer>> instanceSelection;

    public ShiftGenerator(ScheduleConfig config) {

        this.config = config;
        definitions = new ShiftTypeDefinition[config.getShiftTypes()];
        shiftInstances = new ArrayList<>();
        instanceSelection = new ArrayList<>();

    }

    public void generateShifts() {

        for (int i = 0; i < getShiftTypes(); i++) {

            definitions[i] = new ShiftTypeDefinition(config.getShiftSlotLength(), config.getShiftDefinitionTightness());

            int shiftTypeStart = RandomProvider.getInBounds(config.getMinShiftStart(), config.getMaxShiftStart() - (config.getMaxShiftInstances() > 1 ? config.getShiftInstanceDifference() : 0)) * config.getShiftSlotLength();
            int shiftTypeLength = RandomProvider.getInBounds(config.getMinShiftLength(), config.getMaxShiftLength()) * config.getShiftSlotLength();
            int numInstances = RandomProvider.getInBounds(config.getMinShiftInstances(), config.getMaxShiftInstances());

            for (int j = 0; j < numInstances; j++) {

                ShiftInstance newShift;
                do {
                    newShift = new ShiftInstance(i, shiftTypeStart + (j == 0 ? 0 : RandomProvider.get().nextInt(config.getShiftInstanceDifference() + 1) * config.getShiftSlotLength()), getShiftInstanceLength(config, shiftTypeLength));
                } while (checkIdentical(newShift, j));
                shiftInstances.add(newShift);

                definitions[i].addShift(newShift);

            }

            definitions[i].applyTightness();

        }

        generateSelectionList();

    }

    private boolean checkIdentical(ShiftInstance shiftInstance, int count) {

        for (int i = shiftInstances.size() - 1; i >= shiftInstances.size() - count; i--) {

            if (shiftInstances.get(i).getDuration() == shiftInstance.getDuration() && shiftInstances.get(i).getStartTime() == shiftInstance.getStartTime()) {
                return true;
            }

        }

        return false;

    }

    private int getShiftInstanceLength(ScheduleConfig config, int shiftTypeLength) {

        int shiftInstanceLength = shiftTypeLength + RandomProvider.get().nextInt(config.getShiftInstanceDifference() + 1) * config.getShiftSlotLength();

        shiftInstanceLength = shiftInstanceLength > config.getMaxShiftLength() * config.getShiftSlotLength() ?
                Math.max(shiftInstanceLength - config.getShiftInstanceDifference() * config.getShiftSlotLength(), config.getMinShiftLength() * config.getShiftSlotLength()) :
                shiftInstanceLength;

        return shiftInstanceLength;

    }

    private void generateSelectionList() {

        IntStream.range(0, getShiftTypes()).forEach(i -> instanceSelection.add(new ArrayList<>()));
        IntStream.range(0, shiftInstances.size()).forEach(i -> instanceSelection.get(shiftInstances.get(i).getShiftType()).add(i));

    }

    public int getShiftTypes() {
        return definitions.length;
    }

    public List<ShiftInstance> getShiftInstances() {
        return shiftInstances;
    }

    public int getShiftType(int shiftInstanceId) {
        return shiftInstanceId == -1 ? -1 : shiftInstances.get(shiftInstanceId).getShiftType();
    }

    public ShiftTypeDefinition[] getDefinitions() {
        return definitions;
    }

    public boolean checkOverlap(int instance1, int instance2) {

        if (instance1 == -1 || instance2 == -1) {
            return false;
        }

        ShiftInstance inst = shiftInstances.get(instance1);
        return inst.getEndTimeNative() > 24 * 60 && inst.getEndTime() > shiftInstances.get(instance2).getStartTimeNative();

    }

    public List<List<Integer>> getInstanceSelection() {
        return instanceSelection;
    }

}
