package at.ac.tuwien.dbai.ges.instances.shift;

import at.ac.tuwien.dbai.ges.instances.RandomProvider;
import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;
import at.ac.tuwien.dbai.ges.instances.restriction.RestrictionMonitor;

import java.util.*;
import java.util.stream.IntStream;

/**
 * Generates candidates for shift rows and performs the selection.
 *
 * @author Lucas Kletzander
 */
public class ShiftRowGenerator {

    private final ScheduleConfig config;
    private final ShiftGenerator shiftGenerator;
    private final RestrictionMonitor monitor;
    private double[] shiftProb;

    public ShiftRowGenerator(ScheduleConfig config, ShiftGenerator shiftGenerator, RestrictionMonitor monitor) {

        this.config = config;
        this.shiftGenerator = shiftGenerator;
        this.monitor = monitor;
        shiftProb = RandomProvider.initializeRandomArray(shiftGenerator.getShiftTypes(), config.getShiftVariation());

    }

    public void generateRows(int contract, int minOn, int maxOn, int minOff, int maxOff, long amount) {

        double[][] onProb = initOnProb(minOn, maxOn);
        double[][] offProb = initOffProb(minOff, maxOff);

        for (long i = 0; i < amount; i++) {

            int[] assignment = new int[config.getDays() + config.getHistory()];
            Arrays.fill(assignment, -1);

            boolean nextOn = RandomProvider.get().nextDouble() <  ((double) (minOn + maxOn)) / (minOn + maxOn + minOff + maxOff);
            int cur = nextOn ? RandomProvider.getInBounds(-minOn, 0) : RandomProvider.getInBounds(-minOff, 0);

            while (cur < assignment.length) {

                if (nextOn) {

                    int curLength = RandomProvider.getElementFromRandomArray(onProb[Math.floorMod(cur, 7)]) + minOn;
                    IntStream.range(Math.max(cur, 0), Math.min(cur + curLength, assignment.length)).forEach(j -> assignment[j] = chooseShift(assignment, j));
                    cur += curLength;

                } else {

                    cur += RandomProvider.getElementFromRandomArray(offProb[Math.floorMod(cur, 7)]) + minOff;

                }

                nextOn = !nextOn;

            }

            monitor.addRow(new ShiftRow(assignment, contract));

        }

    }

    private double[][] initOnProb(int minOn, int maxOn) {

        double[][] onProb = new double[7][];

        for (int i = 0; i < 7; i++) {

            double[] row = new double[maxOn - minOn + 1];
            Arrays.fill(row, 1);

            for (int j = minOn; j <= maxOn; j++) {

                if ((i + j - 1) % 7 == 4) {
                    row[j - minOn] /= config.getWeekendFactor() < .1 ? .1 : config.getWeekendFactor();
                } else if ((i + j - 1) % 7 > 4) {
                    row[j - minOn] *= Math.pow(config.getWeekendFactor(), 2);
                }

                if ((i + j - 1) / 7 > i / 7) {
                    row[j - minOn] *= Math.sqrt(config.getWeekendFactor());
                }

            }

            onProb[i] = RandomProvider.transformRandomArray(row);

        }

        return onProb;

    }

    private double[][] initOffProb(int minOff, int maxOff) {

        double[][] offProb = new double[7][];

        for (int i = 0; i < 7; i++) {

            double[] row = new double[maxOff - minOff + 1];
            Arrays.fill(row, 1);

            for (int j = minOff; j <= maxOff; j++) {

                if ((i + j - 1) % 7 == 4) {
                    row[j - minOff] *= config.getWeekendFactor();
                } else if ((i + j - 1) % 7 == 5) {
                    row[j - minOff] *= config.getWeekendFactor() == 0 ? 0 : 1;
                } else if ((i + j - 1) % 7 == 6) {
                    row[j - minOff] /= Math.pow(config.getWeekendFactor(), 2) < .01 ? .01 : Math.pow(config.getWeekendFactor(), 2);
                }

                if ((i + j - 1) / 7 > i / 7) {
                    row[j - minOff] /= Math.sqrt(config.getWeekendFactor()) < .01 ? .01 : Math.sqrt(config.getWeekendFactor());
                }

            }

            offProb[i] = RandomProvider.transformRandomArray(row);

        }

        return offProb;

    }

    private int chooseShift(int[] assignment, int position) {

        if (position == 0 || assignment[position - 1] == -1 || RandomProvider.get().nextDouble() > config.getShiftSequenceRegularity()) {

            int shift;
            do {
                shift = RandomProvider.getFromCollection(shiftGenerator.getInstanceSelection().get(RandomProvider.getElementFromRandomArray(shiftProb)));
            } while (hasNeighborCollision(assignment, position, shift));

            return shift;

        } else {
            return assignment[position - 1];
        }

    }

    private boolean hasNeighborCollision(int[] assignment, int day, int shift) {

        return day > 0 && shiftGenerator.checkOverlap(assignment[day - 1], shift) || day < assignment.length - 1 && shiftGenerator.checkOverlap(shift, assignment[day + 1]);

    }

    public List<ShiftRow> chooseShiftRows(int shiftTarget) {

        int[][] dayCount = new int[shiftGenerator.getShiftTypes() + 1][config.getDays() + config.getHistory()];

        int curShifts = 0;
        List<ShiftRow> selection = new ArrayList<>();

        double[] contractProb = RandomProvider.initializeRandomArray(config.getContracts(), config.getContractVariation());

        while (curShifts < shiftTarget) {

            ShiftRow row;

            int contract = RandomProvider.getElementFromRandomArray(contractProb);
            List<ShiftRow> contractRow = monitor.getRows().get(contract);

            PriorityQueue<ShiftRow> candidates = new PriorityQueue<>(Comparator.comparingDouble(ShiftRow::getEvaluation));

            for (int i = 0; i < Math.min(contractRow.size(), (curShifts > shiftTarget * 9 / 10 ? 10000 : 100)); i++) {

                row = RandomProvider.getFromCollection(contractRow);
                row.setEvaluation(evaluateRow(row, dayCount));
                candidates.add(row);

            }

            row = candidates.peek();
            selection.add(row);

            for (int i = 0; i < row.getShifts().length; i++) {

                if (row.getShifts()[i] != -1) {
                    dayCount[shiftGenerator.getShiftType(row.getShifts()[i]) + 1][i]++;
                    dayCount[0][i]++;
                    curShifts++;
                }

            }

        }

        return selection;

    }

    private double evaluateRow(ShiftRow row, int[][] occupation) {

        return IntStream.range(1, occupation.length).mapToDouble(i -> evaluateRow(row, occupation[i], i - 1)).sum();

    }

    private double evaluateRow(ShiftRow row, int[] occupation, int type) {

        double avg = Math.max(IntStream.range(0, occupation.length).filter(i -> i % 7 < 5).mapToDouble(i -> occupation[i]).average().orElse(.5), .5);
        return IntStream.range(0, occupation.length).mapToDouble(i -> addToEvaluation(row, occupation, i, type, avg) ? Math.pow(occupation[i] - getBaseLine(avg, i), 2) * getWeekendWeight(i) : 0).sum();

    }

    private boolean addToEvaluation(ShiftRow row, int[] occupation, int day, int type, double avg) {

        boolean hasType = shiftGenerator.getShiftType(row.getShifts()[day]) == type;
        double difference = occupation[day] - getBaseLine(avg, day);
        return Math.abs(difference) > config.getDayDifference() && (difference > 0 == hasType);

    }

    private double getWeekendWeight(int day) {

        return day % 7 > 4 ? 1 / (config.getWeekendFactor() + .01) : 1;

    }

    private double getBaseLine(double avg, int day) {

        return avg * (day % 7 > 4 ? config.getWeekendFactor() * .95 : 1);

    }

}
