package at.ac.tuwien.dbai.ges.instances.tasks;

import at.ac.tuwien.dbai.ges.instances.RandomProvider;
import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;
import at.ac.tuwien.dbai.ges.instances.breaks.Break;
import at.ac.tuwien.dbai.ges.instances.employee.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftDetails;

import java.util.*;

/**
 * Generates tasks and task schedules.
 *
 * @author Lucas Kletzander
 */
public class TaskGenerator {

    private final ScheduleConfig config;
    private final int[] reAc;
    private final Queue<Task> tasks;
    private final List<Task> commonTasks;

    public TaskGenerator(ScheduleConfig config) {

        this.config = config;
        reAc = new int[config.getTaskTypes()];
        Arrays.setAll(reAc, i -> RandomProvider.get().nextBoolean() ? RandomProvider.get().nextInt(config.getMaxReAcquaintance() + 1) : 0);
        tasks = new PriorityQueue<>(Comparator.comparingInt(Task::getStartGlobalWithWindow));
        commonTasks = new ArrayList<>();

    }

    public void generateTasks(List<EmployeeSchedule> data) {

        for (int i = 0; i < data.size(); i++) {

            EmployeeSchedule schedule = data.get(i);

            for (int j = 0; j < schedule.getDetails().size(); j++) {

                ShiftDetails details = schedule.getDetails().get(j);
                if (details == null || j < config.getHistory()) {
                    continue;
                }

                int curBreak = 0;
                int curTime = details.getStartTime();
                int lastReAc = 0;

                while (curTime < details.getEndTime()) {

                    Task common = checkScheduleCommon(details, j, curBreak, curTime, schedule.getSkills());
                    if (common != null) {

                        addTask(details, j, common);
                        curTime = common.getEnd(j);
                        curBreak = updateBreakIndex(details, curBreak, curTime);
                        continue;

                    }

                    if (curTime == details.getStartTime()) {
                        curTime += RandomProvider.getInBounds(0, config.getShiftSlotLength() - config.getTimeSlotLength(), config.getTimeSlotLength());
                    } else {
                        int newStart = curTime + (RandomProvider.get().nextBoolean() ? RandomProvider.get().nextInt(Math.max(config.getTaskDistanceRange() - lastReAc + 1, 1)) * config.getTimeSlotLength() : 0);
                        if (newStart <= getLatestStart(details, curBreak)) {
                            curTime = newStart;
                        }
                    }

                    curBreak = updateBreakIndex(details, curBreak, curTime);
                    if (curBreak < details.getBreaks().size() && details.getBreaks().get(curBreak).getStart() <= curTime) {
                        curTime = details.getBreaks().get(curBreak).getEnd();
                        curBreak = updateBreakIndex(details, curBreak, curTime);
                    }

                    int type = RandomProvider.get().nextInt(reAc.length);

                    if (curBreak < details.getBreaks().size() && curTime + config.getMinTaskLength() * config.getTimeSlotLength() > details.getBreaks().get(curBreak).getStart()) {
                        if (config.getTaskGenerationType() == TaskGenerationType.NO_BREAKS) {
                            curTime += config.getTimeSlotLength();
                            continue;
                        } else if (type >= config.getTaskGenerationType().ordinal()) {
                            type = RandomProvider.get().nextInt(config.getTaskGenerationType().ordinal());
                        }
                    }

                    int length;
                    if (type == config.getTaskGenerationType().ordinal() - 1) {
                        length = RandomProvider.getInBounds(config.getMinTaskLength(), config.getMaxTaskLength()) * config.getTimeSlotLength();
                    } else if (type == config.getTaskGenerationType().ordinal() - 2) {
                        length = RandomProvider.getInBounds(config.getMinTaskLength(), config.getMaxTaskLength() * 3) * config.getTimeSlotLength();
                    } else if (type == config.getTaskGenerationType().ordinal() - 3) {
                        length = RandomProvider.getInBounds(config.getMinTaskLength(), config.getMaxTaskLength() / 2) * config.getTimeSlotLength();
                    } else {
                        length = RandomProvider.getInBounds(config.getMinTaskLength() * config.getTimeSlotLength(),
                                Math.min(config.getMaxTaskLength() * config.getTimeSlotLength(),
                                        (curBreak < details.getBreaks().size() ? details.getBreaks().get(curBreak).getStart() : Integer.MAX_VALUE) - curTime),
                                config.getTimeSlotLength());
                    }

                    Task task = new Task(type, RandomProvider.getFromCollection(schedule.getSkills(), -1));
                    curTime = setTaskTimes(details, j, task, curBreak, curTime, length);

                    if (task.getEnd(j) <= details.getEndTime()) {
                        addTask(details, j, task);
                    }

                    lastReAc = reAc[type];
                    curTime += lastReAc * config.getTimeSlotLength();

                }

                if (i < data.size() * config.getCommonTaskFactor() * .2 + 1 && RandomProvider.get().nextDouble() < config.getCommonTaskFactor()) {
                    Task t = RandomProvider.getFromCollection(details.getTasks());
                    if (t.getType() >= config.getTaskGenerationType().ordinal()) {
                        commonTasks.add(t);
                    }
                }

                unifySplitTask(details.getTasks(), j);

            }

        }

    }

    private int getLatestStart(ShiftDetails details, int curBreak) {

        if (curBreak < details.getBreaks().size()) {
            return details.getBreaks().get(curBreak).getStart() - config.getMinTaskLength() * config.getTimeSlotLength();
        } else {
            return details.getEndTime() - config.getMinTaskLength() * config.getTimeSlotLength();
        }

    }

    private int updateBreakIndex(ShiftDetails details, int index, int time) {

        while (index < details.getBreaks().size() && details.getBreaks().get(index).getEnd() <= time) {
            index++;
        }

        return index;

    }

    private int setTaskTimes(ShiftDetails details, int day, Task task, int curBreak, int curTime, int length) {

        Break br;
        while (curBreak < details.getBreaks().size() && (br = details.getBreaks().get(curBreak)).getStart() < curTime + length) {
            task.add(day, curTime, br.getStart());
            length -= br.getStart() - curTime;
            curTime = br.getEnd();
            curBreak++;
        }

        if (curTime + length > details.getEndTime() && curTime + config.getMinTaskLength() * config.getTimeSlotLength() <= details.getEndTime()) {
            task.add(day, curTime, RandomProvider.getInBounds(Math.max(curTime + config.getMinTaskLength() * config.getTimeSlotLength(), details.getEndTime() - config.getShiftSlotLength() + config.getTimeSlotLength()), details.getEndTime(), config.getTimeSlotLength()));
        } else {
            task.add(day, curTime, curTime + length);
        }

        return curTime + length;

    }

    private Task checkScheduleCommon(ShiftDetails details, int day, int curBreak, int curTime, List<Integer> skills) {

        int globalTime = day * 24 * 60 + curTime;

        for (Task t : commonTasks) {

            if (skills.contains(t.getSkill()) && t.getStartGlobal() >= globalTime && t.getStartGlobal() <= globalTime + config.getTaskDistanceRange() * config.getTimeSlotLength()) {

                int newBreak = updateBreakIndex(details, curBreak, t.getStart(day));
                if ((newBreak >= details.getBreaks().size() && t.getEnd(day) <= details.getEndTime()) || (newBreak < details.getBreaks().size() && t.getEnd(day) <= details.getBreaks().get(newBreak).getStart())) {
                    return t;
                }

            }

        }

        return null;

    }

    private void addTask(ShiftDetails details, int day, Task task) {

        details.getTasks().add(task);

        int timeWindow = RandomProvider.get().nextInt(config.getTaskWindowRange() + 1) * config.getTimeSlotLength();
        int timeWindowStart = RandomProvider.getInBounds(0, timeWindow, config.getTimeSlotLength());
        task.setWindow(timeWindowStart, timeWindow - timeWindowStart);

        if (day >= config.getHistory() && task.addUsage()) {
            tasks.add(task);
        }

    }

    private void unifySplitTask(List<Task> list, int day) {

        Task current = null;
        Iterator<Task> it = list.iterator();

        while (it.hasNext()) {

            Task task = it.next();

            if (task.getType() == config.getTaskGenerationType().ordinal() - 3) {

                if (current == null || task.getEnd(day) > current.getStart(day) + config.getMaxTaskLength() * config.getTimeSlotLength() * 5) {
                    current = task;
                } else {

                    for (int i = 0; i < task.getStartList().size(); i++) {
                        current.add(task.getDay(), task.getStartList().get(i), task.getEndList().get(i));
                    }
                    it.remove();
                    tasks.remove(task);

                }

            }

        }

    }

    public int[] getReAc() {
        return reAc;
    }

    public Queue<Task> getTasks() {
        return tasks;
    }

}
