package at.ac.tuwien.dbai.ges.instances.transformer;

import at.ac.tuwien.dbai.ges.instances.RandomProvider;
import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;
import at.ac.tuwien.dbai.ges.instances.breaks.*;
import at.ac.tuwien.dbai.ges.schema.BreakConfigurations;
import at.ac.tuwien.dbai.ges.schema.BreakTypes;
import at.ac.tuwien.dbai.ges.schema.Breaks;
import at.ac.tuwien.dbai.ges.schema.ObjectFactory;

import javax.xml.bind.JAXBElement;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Arrays;
import java.util.List;

/**
 * Transforms the break section of the problem to XML.
 *
 * @author Lucas Kletzander
 */
public class BreakTransformer {

    private final ScheduleConfig config;
    private final ObjectFactory objectFactory;
    private int typeCount = 0;

    public BreakTransformer(ScheduleConfig config, ObjectFactory objectFactory) {

        this.config = config;
        this.objectFactory = objectFactory;

    }

    public Breaks transform(BreakGenerator breakGenerator) {

        if (breakGenerator == null) {
            return null;
        }

        Breaks breaks = new Breaks();

        BreakTypes types = new BreakTypes();
        List<BreakTypes.BreakType> typeList = types.getBreakType();
        Arrays.stream(breakGenerator.getBreakTypes()).forEach(i -> typeList.add(transformBreakType(i)));
        breaks.setBreakTypes(types);

        BreakConfigurations configs = new BreakConfigurations();
        List<BreakConfigurations.BreakConfiguration> configList = configs.getBreakConfiguration();
        breakGenerator.getBreakConfigurations().forEach(i -> configList.add(transformBreakConfiguration(breakGenerator.getBreakTypes(), i)));
        breaks.setBreakConfigurations(configs);

        return breaks;

    }

    private BreakTypes.BreakType transformBreakType(BreakType type) {

        BreakTypes.BreakType t = new BreakTypes.BreakType();
        t.setID("B" + typeCount);
        typeCount++;

        List<JAXBElement<?>> data = t.getMinLengthOrMaxLengthOrMinStart();
        data.add(objectFactory.createBreakTypesBreakTypeMinLength(Transformer.transformTimeSpan(type.getMinLength())));
        data.add(objectFactory.createBreakTypesBreakTypeMaxLength(Transformer.transformTimeSpan(type.getMaxLength())));

        if (type.getStartShiftWeight() > 0) {
            data.add(objectFactory.createBreakTypesBreakTypeMinStartShift(Transformer.transformTimeSpan(type.getMinStartShift(), type.getStartShiftWeight())));
            data.add(objectFactory.createBreakTypesBreakTypeMaxStartShift(Transformer.transformTimeSpan(type.getMaxStartShift(), type.getStartShiftWeight())));
        }

        data.add(objectFactory.createBreakTypesBreakTypeMaxEndShift(Transformer.transformTimeSpan(type.getMaxEndShift())));
        data.add(objectFactory.createBreakTypesBreakTypeMinWorkBefore(String.valueOf(type.getMinWorkBefore())));
        data.add(objectFactory.createBreakTypesBreakTypeMaxWorkBefore(String.valueOf(type.getMaxWorkBefore())));

        return t;

    }

    private BreakConfigurations.BreakConfiguration transformBreakConfiguration(BreakType[] breakTypes, BreakConfiguration breakConfig) {

        BreakConfigurations.BreakConfiguration c = new BreakConfigurations.BreakConfiguration();

        if (config.getBreakConfigurationType() != BreakConfigurationType.COMMON) {

            BreakConfigurations.BreakConfiguration.ShiftFilter shiftFilter = new BreakConfigurations.BreakConfiguration.ShiftFilter();

            switch (config.getBreakConfigurationType()) {

                case PER_SHIFT_TYPE:
                    shiftFilter.setShiftTypes("S" + breakConfig.getId());
                    break;

                case PER_CONTRACT:
                    BreakConfigurations.BreakConfiguration.ShiftFilter.Contracts contracts = new BreakConfigurations.BreakConfiguration.ShiftFilter.Contracts();
                    contracts.getContract().add("C" + breakConfig.getId());
                    shiftFilter.setContracts(contracts);
                    break;

                case PER_SHIFT_LENGTH:
                    shiftFilter.setMinShiftLength(String.valueOf(breakConfig.getMinLength()));
                    shiftFilter.setMaxShiftLength(String.valueOf(breakConfig.getMaxLength()));
                    break;

            }

            c.setShiftFilter(shiftFilter);

        }

        if (breakConfig.getMinBreakTimeFraction() > breakConfig.getMaxBreakTimeFraction()) {
            breakConfig.addBreakTimeFraction(RandomProvider.get().nextDouble() * .1 + .05);
            breakConfig.addBreakTimeFraction(RandomProvider.get().nextDouble() * .1 + .15);
        }

        if (config.getBreakTimeFractionConstraintType() == BreakTimeFractionConstraintType.MIN || config.getBreakTimeFractionConstraintType() == BreakTimeFractionConstraintType.MIN_MAX) {
            c.setMinTotalBreakTimeFraction(Transformer.transformDoubleRound(breakConfig.getMinBreakTimeFraction(), BigDecimal.ROUND_DOWN));
        }

        if (config.getBreakTimeFractionConstraintType() == BreakTimeFractionConstraintType.MAX || config.getBreakTimeFractionConstraintType() == BreakTimeFractionConstraintType.MIN_MAX) {
            c.setMaxTotalBreakTimeFraction(Transformer.transformDoubleRound(breakConfig.getMaxBreakTimeFraction(), BigDecimal.ROUND_UP));
        }

        c.setMaxWorkAfterLast(HorizonTransformer.transformTimeSpan(breakConfig.getMaxWorkAfterLast()));

        BreakConfigurations.BreakConfiguration.BreakSet breakSet = new BreakConfigurations.BreakConfiguration.BreakSet();
        List<BreakConfigurations.BreakConfiguration.BreakSet.Break> breakList = breakSet.getBreak();

        for (Integer id : breakConfig.getAvailableTypes()) {

            BreakConfigurations.BreakConfiguration.BreakSet.Break b = new BreakConfigurations.BreakConfiguration.BreakSet.Break();

            if (breakTypes[id].getStartShiftWeight() > 0) {
                b.setMaxCount(BigInteger.ONE);
                if (!breakConfig.getTypeUnused()[id]) {
                    b.setMinCount(BigInteger.ONE);
                }
            }

            b.setValue("B" + id);
            breakList.add(b);

        }

        c.setBreakSet(breakSet);

        return c;

    }

}
