package at.ac.tuwien.dbai.ges.instances.transformer;

import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;
import at.ac.tuwien.dbai.ges.instances.employee.EmployeeMonitor;
import at.ac.tuwien.dbai.ges.instances.employee.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.instances.restriction.Restriction;
import at.ac.tuwien.dbai.ges.schema.*;

import javax.xml.bind.JAXBElement;
import java.util.List;
import java.util.stream.IntStream;

/**
 * Transforms the employee section of the problem to XML.
 *
 * @author Lucas Kletzander
 */
public class EmployeeTransformer {

    private final ScheduleConfig config;
    private final ObjectFactory objectFactory;

    public EmployeeTransformer(ScheduleConfig config, ObjectFactory objectFactory) {

        this.config = config;
        this.objectFactory = objectFactory;

    }

    public Employees transform(List<EmployeeSchedule> rows, List<Restriction> restrictions, EmployeeMonitor monitor, int skills) {

        Employees employees = new Employees();
        Contracts contracts = new Contracts();

        for (int i = 0; i < config.getContracts(); i++) {

            Contracts.Contract contract = new Contracts.Contract();
            contract.setID("C" + i);
            contract.setLabel("C" + i);

            transformContract(contract, i);

            for (Restriction restriction : restrictions) {
                restriction.transform(objectFactory, contract);
            }

            contracts.getContract().add(contract);

        }

        employees.setContracts(contracts);

        employees.setSkills(transformSkills(skills));

        EmployeeList employeeList = new EmployeeList();
        List<AbstractEmployeeType> eList = employeeList.getEmployeeOrVariableEmployee();

        for (int i = 0; i < rows.size(); i++) {

            EmployeeList.Employee employee = new EmployeeList.Employee();
            employee.setID("E" + i);
            employee.setName("Employee " + i);
            employee.setContracts(transformEmployeeContracts(i, contracts, rows.get(i)));

            monitor.applyConfigurations(employee, rows.get(i));

            if (employee.getPreAssignments().getShiftOrNoShift().isEmpty()) {
                employee.setPreAssignments(null);
            }

            eList.add(employee);

        }

        employees.setEmployeeList(employeeList);
        return employees;

    }

    private void transformContract(Contracts.Contract c, int contract) {

        Contracts.Contract.PatternConstraints p = c.getPatternConstraints();
        if (p == null) {
            p = new Contracts.Contract.PatternConstraints();
            c.setPatternConstraints(p);
        }

        List<JAXBElement<?>> list = p.getCountConstraintOrSequenceConstraintOrWeekendCount();
        list.add(getSequenceConstraint(false, config.getMinOff()[contract], config.getMaxOff()[contract]));
        list.add(getSequenceConstraint(true, config.getMinOn()[contract], config.getMaxOn()[contract]));

    }

    private JAXBElement<SequenceCount> getSequenceConstraint(boolean on, int min, int max) {

        SequenceCount seq = new SequenceCount();

        if (on) {
            seq.setAnyShift(new AnyShift());
        } else {
            seq.setNoShift(new NoShift());
        }

        List<JAXBElement<WeightedInteger>> minMax = seq.getMinOrMax();
        minMax.add(objectFactory.createSequenceCountMin(Transformer.transformInteger(min)));
        minMax.add(objectFactory.createSequenceCountMax(Transformer.transformInteger(max)));
        return objectFactory.createContractsContractPatternConstraintsSequenceConstraint(seq);

    }

    private AbstractEmployeeType.Contracts transformEmployeeContracts(int id, Contracts contracts, EmployeeSchedule schedule) {

        AbstractEmployeeType.Contracts eContracts = new AbstractEmployeeType.Contracts();
        eContracts.getContract().add("C" + schedule.getContract());

        if (schedule.getSkills().size() > 0) {
            Contracts.Contract personal = new Contracts.Contract();
            personal.setID("CP" + id);
            Contracts.Contract.ProvidedSkills skills = new Contracts.Contract.ProvidedSkills();
            schedule.getSkills().forEach(i -> skills.getSkill().add("Sk" + i));
            personal.setProvidedSkills(skills);
            contracts.getContract().add(personal);
            eContracts.getContract().add("CP" + id);
        }

        return eContracts;

    }

    private Skills transformSkills(int amount) {

        if (amount == 0) {
            return null;
        }

        Skills skills = new Skills();
        IntStream.range(0, amount).forEach(i -> {
            Skills.Skill skill = new Skills.Skill();
            skill.setID("Sk" + i);
            skills.getSkill().add(skill);
        });

        return skills;

    }

}
