package at.ac.tuwien.dbai.ges.instances.transformer;

import at.ac.tuwien.dbai.ges.instances.RandomProvider;
import at.ac.tuwien.dbai.ges.instances.ScheduleConfig;
import at.ac.tuwien.dbai.ges.instances.employee.EmployeeSchedule;
import at.ac.tuwien.dbai.ges.instances.shift.ShiftDetails;
import at.ac.tuwien.dbai.ges.schema.CoverRequirements;
import at.ac.tuwien.dbai.ges.schema.Demands;
import at.ac.tuwien.dbai.ges.schema.ObjectFactory;
import at.ac.tuwien.dbai.ges.schema.TimeCover;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 * Transforms the problem to time-based demand specification.
 *
 * @author Lucas Kletzander
 */
public class TimeTransformer {

    private final ScheduleConfig config;
    private final ObjectFactory objectFactory;

    public TimeTransformer(ScheduleConfig config, ObjectFactory objectFactory) {
        this.config = config;
        this.objectFactory = objectFactory;
    }

    public Demands transformDemands(List<EmployeeSchedule> schedules) {

        int[] count = new int[config.getDays() * 24 * 60 / config.getTimeSlotLength()];
        List<Set<Integer>> skills = new ArrayList<>();
        IntStream.range(0, count.length).forEach(i -> skills.add(new HashSet<>()));

        for (EmployeeSchedule schedule : schedules) {

            for (int i = 0; i < schedule.getDetails().size() - config.getHistory(); i++) {

                ShiftDetails details = schedule.getDetails().get(i + config.getHistory());

                if (details != null) {

                    int breakIndex = 0;

                    for (int j = details.getStartTime(); j < details.getEndTime(); j += config.getTimeSlotLength()) {

                        if ((i * 24 * 60 + j) / config.getTimeSlotLength() >= count.length) {
                            break;
                        }

                        while (breakIndex < details.getBreaks().size() && details.getBreaks().get(breakIndex).getEnd() <= j) {
                            breakIndex++;
                        }

                        if (breakIndex == details.getBreaks().size() || j < details.getBreaks().get(breakIndex).getStart()) {

                            int index = (i * 24 * 60 + j) / config.getTimeSlotLength();
                            count[index]++;
                            skills.get(index).addAll(schedule.getSkills());

                        }

                    }
                }

            }

        }

        Demands demands = new Demands();
        int timeSlotsPerBlock = config.getDemandTimeBlock() == -1 ? 1 : config.getDemandTimeBlock() / config.getTimeSlotLength();
        Set<Integer> curSkills = new HashSet<>();

        for (int i = 0; i < count.length; i += timeSlotsPerBlock) {

            int min = Arrays.stream(count, i, i + timeSlotsPerBlock).min().orElse(0);
            int max = Arrays.stream(count, i, i + timeSlotsPerBlock).max().orElse(0);
            Set<Integer> activeSkills = skills.get(i);
            IntStream.range(i + 1, i + timeSlotsPerBlock).forEach(j -> activeSkills.retainAll(skills.get(j)));

            while (curSkills.size() > max) {
                curSkills.remove(RandomProvider.getFromCollection(curSkills));
            }
            curSkills.removeAll(curSkills.stream().filter(j -> RandomProvider.get().nextDouble() < .05).collect(Collectors.toList()));

            while (curSkills.size() < max && curSkills.size() < config.getDemandSkills() && curSkills.size() < activeSkills.size() && RandomProvider.get().nextDouble() < (curSkills.size() == 0 ? .8 : .2)) {
                curSkills.add(RandomProvider.getFromCollection(activeSkills));
            }

            if (max == 0) {
                continue;
            }

            TimeCover cover = new TimeCover();
            cover.setID("TC" + i);
            cover.setDay(String.valueOf(i * config.getTimeSlotLength() / (24 * 60)));
            cover.setStartTime(Transformer.transformTimePoint(i * config.getTimeSlotLength() % (24 * 60)));
            cover.setEndTime(Transformer.transformTimePoint((i + timeSlotsPerBlock) * config.getTimeSlotLength() % (24 * 60)));

            CoverRequirements requirements = new CoverRequirements();

            curSkills.forEach(j -> requirements.getMinOrMax().add(objectFactory.createCoverRequirementsMin(Transformer.transformIntSkill(1, j))));

            int minSlack = (int) (RandomProvider.get().nextDouble() * config.getDemandSlack() * min);
            requirements.getMinOrMax().add(objectFactory.createCoverRequirementsMin(Transformer.transformIntSkill(Math.max(min - minSlack, 0))));

            if (config.isDemandUseMax()) {
                requirements.getMinOrMax().add(objectFactory.createCoverRequirementsMax(Transformer.transformIntSkill(max + (int) (RandomProvider.get().nextDouble() * config.getDemandSlack() * (min + 1)))));
            }

            cover.setRequirements(requirements);
            demands.getTimeCover().add(cover);

        }

        return demands;

    }

}
